@@ -83,7 +83,7 @@ Return a fully qualified incomplete Eta instance
 =cut
 sub freshEta {
     my ($this) = @_;
-    return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Eta->new($this->freshKappa, $this->genOmicron, $this->genChi));
+    return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Eta->new($this->freshKappa, $this->genOmicron, $this->genChi, $this->freshZeta));
 }
 
 =item B<freshNu>
@@ -116,6 +116,16 @@ sub freshUpsilon {
     return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Kappa->new(Devel::TypeCheck::Type::Upsilon->new($this->fresh())));
 }
 
+=item B<freshZeta>
+
+Return a fully qualified incomplete Zeta instance
+
+=cut
+sub freshZeta {
+    my ($this) = @_;
+    return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Zeta->new($this->genOmicron, $this->fresh));
+}
+
 =item B<genRho>
 
 Encapsulate something in a fully qualified reference
@@ -182,6 +192,16 @@ sub genChi {
     return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Chi->new($subtype));
 }
 
+=item B<genZeta>($params, $return)
+
+Generate a Zeta type with the given params and return value.
+
+=cut
+sub genZeta {
+    my ($this, $params, $return) = @_;
+    return Devel::TypeCheck::Type::Mu->new(Devel::TypeCheck::Type::Zeta->new($params, $return));
+}
+
 # Union two types, as per union-find data structure
 sub union {
     my ($this, $t1, $t2) = @_;
